import weakref

import aioamqp
from aioamqp.channel import Channel
import asyncio

from .transportabc import TransportABC, ClientTransportABC
from ..webtypes import Request, Response


class RabbitMQClientTransport(ClientTransportABC):
    def __init__(self):
        raise NotImplementedError

    def make_request(self, service: str, method: str, path: str,
                     body: bytes = None, query: str = None,
                     headers: dict = None, correlation_id: str = None,
                     content_type: str = None, **kwargs):
        raise NotImplementedError


class RabbitMQTransport(TransportABC):
    def __init__(self, *, url, port=5672, queue='', virtualhost='/',
                 username='guest', password='guest',
                 ssl=False, verify_ssl=True):
        self.host = url
        self.port = port
        self.virtualhost = virtualhost
        self.queue = queue
        self.username = username
        self.password = password
        self.ssl = ssl
        self.verify_ssl = verify_ssl
        self._transport = None
        self._protocol = None
        self.channel = None
        self._app = None
        self._loop = None
        self._consumer_tag = None
        self._counter = 0
        self._handler = None

    def get_client(self):
        pass

    async def declare_exchange(self):
        pass

    async def declare_queue(self):
        pass

    async def bind_to_exchange(self, *, exchange, routing_key):
        await self.channel.queue_bind(exchange_name=exchange,
                                      queue_name=self.queue,
                                      routing_key=routing_key)

    def start(self, handler, *, loop):
        self._handler = handler
        self._loop = loop
        async def consume():
            # Need to reconnect because of potential forking affects
            await self.close()
            await self.connect(loop=loop)
            self._consumer_tag = (await self.channel.basic_consume(
                self.handle_request,
                queue_name=self.queue,
                no_ack=True)).get('consumer_tag')

        loop.run_until_complete(consume())

    def listen(self, *, loop):
        async def setup():

            await self.connect(loop=loop)
            await self.channel.queue_declare(queue_name=self.queue)
            await self.channel.basic_qos(prefetch_count=1, prefetch_size=0)

        loop.run_until_complete(setup())
        print('-- Listening for rabbitmq messages on queue {} --'
              .format(self.queue))

    async def connect(self, loop=None):
        self._transport, self._protocol = await aioamqp.connect(
            host=self.host,
            port=self.port,
            virtualhost=self.virtualhost,
            login=self.username,
            password=self.password,
            ssl=self.ssl,
            verify_ssl=self.verify_ssl,
            loop=loop
        )
        self.channel = await self._protocol.channel()

    async def close(self):
        await self._protocol.close()
        self._transport.close()

    async def handle_request(self, channel: Channel, body, envelope,
                             properties, futurize=True):
        """
        the 'f' param is simply because aioamqp doesnt send another job until
         this method returns (completes), so we ensure the future of
         ourselves and return immediately so we can handle many requests
         at a time.
        """
        if futurize:
            asyncio.ensure_future(
                self.handle_request(channel, body, envelope, properties,
                                    futurize=False))
            return

        self._counter += 1
        route = envelope.routing_key
        print(route)
        headers = properties.headers or {}
        query = headers.pop('query', '').lstrip('?')
        headers['content-type'] = properties.content_type
        headers['content-encoding'] = properties.content_encoding
        correlation_id = properties.correlation_id
        reply_to = properties.reply_to
        method = properties.type

        request = Request(
            headers=headers,
            path=route,
            correlation_id=correlation_id,
            method=method,
            query_string=query,
            body=body,
        )

        response = await self._handler(request)
        if not reply_to:
            # nowhere to reply. No point it trying to send a response
            return

        response.headers['Status'] = str(response.status.value)

        properties = {
            'correlation_id': response.correlation_id,
            'headers': response.headers,
            'content_type': response.content_type
        }
        await channel.basic_publish(response.data, '', reply_to,
                                    properties=properties)
        self._counter -= 1

    def shutdown(self, *, loop):
        loop.run_until_complete(self.channel.basic_cancel(self._consumer_tag))
        async def finish_tasks():
            while self._counter > 0:
                await asyncio.sleep(1)
        loop.run_until_complete(finish_tasks())
        loop.run_until_complete(self.close())





from enum import Enum

from waspy import webtypes


class NonRESTfulURLError(NotImplementedError):
    """ URL provided is not restful. """


class Methods(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    HEAD = 'HEAD'


def _parameterize_path(path):
    path_parts = path.split('.')

    params = []
    on_id_segment = False
    key = ''
    for part in path_parts:
        if on_id_segment:
            # this part is a resource id
            key += '.*'
            params.append(part)
        else:
            # this part is a resource name
            key += '.' + part
        on_id_segment = not on_id_segment
    return key, params


async def _send_404(request):
    raise webtypes.ResponseError(status=404)


class Router:
    def __init__(self):
        self._routes = {}
        """
        Routes looks like:
        {method: {path: (handler, params)}} before startup and
        {method: {path: (wrapped, handler, params)}} after startup
        """
        self._static_routes = {}
        """
        Static routes look similar to routes, but with only a handler, no tuple
        (before the app is started, they also include a boolean to whether
        they should be wrapped or not {method: {path: handler, True}})
        {method: {path: handler}}
        """
        self._no_wrap_static_routes = {}
        self.options_handler = None

        self.handle_404 = _send_404  # the 404 route will skip middlewares

    def _get_and_wrap_routes(self):
        for method, routes in self._routes.items():
            for path, t in routes.items():
                handler, params = t
                wrapped = yield handler
                routes[path] = (wrapped, handler, params)
        for method, routes in self._static_routes.items():
            for path, t in routes.items():
                handler, should_not_wrap = t
                if should_not_wrap:
                    wrapped = handler
                else:
                    wrapped = yield handler
                routes[path] = wrapped

    def _prepare_route(self, route):
        route = route.replace('/', '.').lstrip('.')
        key, params = _parameterize_path(route)

        # Check that route is RESTful
        for p in params:
            if (not p.startswith(':') or
                    (not p.startswith('{') and p.endswith('}'))):
                # Param is not an id parameter
                raise NonRESTfulURLError('Path variable expected in url, '
                                         'instead got {}')
        if ':' in key or '{' in key:
            raise NonRESTfulURLError('Got a path variable in URL when a '
                                     'resource name was expected')

        params = [p.strip(':').strip('{}') for p in params]
        return key, params

    def get_handler_for_request(self, request):
        method = request.method
        path = request.path
        route = path.lstrip('/').replace('/', '.')
        if method == Methods.OPTIONS and self.options_handler is not None:
            request._handler = self.options_handler
            return self.options_handler

        try:
            return self._static_routes[method][route]
        except KeyError:
            # not in static routes
            pass

        path, params = _parameterize_path(route)
        try:
            wrapped, handler, keys = self._routes[method][path]
        except KeyError:
            request._handler = self.handle_404
            return self.handle_404
        assert len(keys) == len(params)
        for key, param in zip(keys, params):
            request.path_params[key] = param
        request._handler = handler
        return wrapped

    def add_static_route(self, method: str, route: str, handler: callable,
                         skip_middleware=False):
        """
        Adds a static route. A static route is a special route that
        doesnt follow any of the normal rules, and never has any path
        parameters.
        Ideally, this is used for non-public facing endpoints such as
        "/healthcheck", or "/stats" or something of that nature.

        Set `skip_middleware` to true to bypass middlewares
        """
        if not isinstance(method, Methods):
            method = Methods(method.upper())
        if method not in self._static_routes:
            self._static_routes[method] = {}
        route = route.lstrip('/').replace('/', '.')
        self._static_routes[method][route] = handler, skip_middleware

    def add_route(self, method: str, route: str, handler: callable):
        if not isinstance(method, Methods):
            method = Methods(method.upper())
        if method not in self._routes:
            self._routes[method] = {}
        prepared_route, params = self._prepare_route(route)
        self._routes[method][prepared_route] = handler, params

    def add_get(self, route: str, handler: callable):
        self.add_route(Methods.GET, route, handler)

    def add_post(self, route: str, handler: callable):
        self.add_route(Methods.POST, route, handler)

    def add_put(self, route: str, handler: callable):
        self.add_route(Methods.PUT, route, handler)

    def add_delete(self, route: str, handler: callable):
        self.add_route(Methods.DELETE, route, handler)

    def add_patch(self, route: str, handler: callable):
        self.add_route(Methods.PATCH, route, handler)

    def add_head(self, route: str, handler: callable):
        self.add_route(Methods.HEAD, route, handler)

    def add_options(self, route: str, handler: callable):
        self.add_route(Methods.OPTIONS, route, handler)

    def add_generic_options_handler(self, handler: callable):
        """
        Add a handler for all options requests. This WILL bypass middlewares
        """
        self.options_handler = handler



import asyncio
import socket

import h11

from ..webtypes import Request, Response
from .transportabc import TransportABC, ClientTransportABC


class ClosedError(Exception):
    """ Error for closed connections """


class _HTTPClientConnection:
    slots = ('reader', 'writer', 'conn', 'port', 'service')

    def __init__(self, service, port):
        self.reader = None
        self.writer = None
        self.port = port
        self.service = service
        self.conn = h11.Connection(our_role=h11.CLIENT)

    async def connect(self):
        self.reader, self.writer = await \
            asyncio.open_connection(self.service, self.port)

    def send(self, event):
        data = self.conn.send(event)
        if data is not None:
            self.writer.write(data)

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                data = await self.reader.read(1064)
                self.conn.receive_data(data)

            else:
                return event

    def close(self):
        self.send(h11.ConnectionClosed())
        self.writer.close()


class HTTPClientTransport(ClientTransportABC):
    """Client implementation of the HTTP transport protocol"""
    def _get_connection_for_service(self, service):
        pass

    async def make_request(self, service, method, path, body=None, query=None,
                           headers=None, correlation_id=None,
                           content_type=None, port=80, **kwargs):
        # form request object
        path = path.replace('.', '/')
        if headers is None:
            headers = {}
        headers['Host'] = service
        headers['Connection'] = 'close'
        if correlation_id:
            headers['X-Correlation-Id'] = correlation_id
        if query:
            path += '?' + query
        if content_type:
            headers['Content-Type'] = content_type
        if body:
            headers['Content-Length'] = str(len(body))

        # now make a connection and send it
        connection = _HTTPClientConnection(service, port)
        await connection.connect()
        connection.send(h11.Request(method=method, target=path,
                                    headers=headers.items()))
        if body:
            connection.send(h11.Data(data=body))

        connection.send(h11.EndOfMessage())

        response = await connection.next_event()
        assert type(response) is h11.Response

        # form response object
        status_code = response.status_code
        headers = response.headers

        result = Response(headers=headers, correlation_id=correlation_id,
                          status=status_code)

        body = b''
        event = None
        while type(event) is not h11.EndOfMessage:
            event = await connection.next_event()
            if type(event) is h11.Data:
                body += body

        result.body = body
        connection.close()
        return result


class HTTPTransport(TransportABC):
    """ Server implementation of the HTTP transport protocol"""

    def get_client(self):
        return HTTPClientTransport()

    def __init__(self, port=8080, prefix=None):
        self.port = port
        if prefix is None:
            prefix = ''
        self.prefix = prefix
        self._handler = None
        self._server = None

    def listen(self, *, loop: asyncio.AbstractEventLoop):
        coro = asyncio.start_server(
            self.handle_incoming_request, '0.0.0.0', self.port, loop=loop)
        self._server = loop.run_until_complete(coro)
        print('-- Listening for HTTP on port {} --'.format(self.port))

    def start(self, request_handler, *, loop: asyncio.AbstractEventLoop):
        self._handler = request_handler

    async def handle_incoming_request(self, reader, writer):
        self._set_tcp_nodelay(writer)
        protocol = _HTTPServerProtocol(reader=reader, writer=writer,
                                       prefix=self.prefix)

        try:
            while True:
                request = await protocol.get_request()
                response = await self._handler(request)
                await protocol.send_response(response)
                if protocol.conn.our_state is h11.MUST_CLOSE:
                    break
                protocol.conn.start_next_cycle()
        except (ClosedError, ConnectionResetError):
            pass
        writer.close()

    def shutdown(self, *, loop):
        self._server.close()
        loop.run_until_complete(self._server.wait_closed())

    def _set_tcp_nodelay(self, writer):
        socket_ = writer._transport._sock
        if socket_ is None:
            return
        if socket_.family not in (socket.AF_INET, socket.AF_INET6):
            return
        socket_.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)


class _HTTPServerProtocol(asyncio.Protocol):
    """ HTTP Protocol handler.
        Should only be used by HTTPServerTransport
    """
    __slots__ = ('conn', 'reader', 'writer')

    def __init__(self, reader=None, writer=None, prefix=None):
        self.conn = h11.Connection(h11.SERVER)
        self.reader = reader
        self.writer = writer
        self.prefix = prefix

    async def _read(self):
        if self.conn.they_are_waiting_for_100_continue:
            go_ahead = h11.InformationalResponse(status_code=100)
            await self.writer.write(go_ahead)
        try:
            data = await self.reader.read(1000000)
        except ConnectionError:
            data = b''
        self.conn.receive_data(data)

    async def send_response(self, response: Response):
        headers = [(name, value) for name, value in response.headers.items()]
        r = h11.Response(status_code=response.status.value, headers=headers,
                         reason=response.status.phrase)
        await self._send(r)
        await self._send(h11.Data(data=response.data))
        await self._send(h11.EndOfMessage())

    async def _send(self, r):
        data = self.conn.send(r)
        self.writer.write(data)
        await self.writer.drain()

    async def get_request(self) -> Request:

        event = await self.next_event()
        if isinstance(event, h11.ConnectionClosed):
            raise ClosedError
        headers = {name.decode('ascii'): value.decode('ascii')
                   for name, value in event.headers}

        # split path and query string
        target = event.target.decode('ascii').split('?', maxsplit=1)
        path = target[0].lstrip(self.prefix)
        try:
            query = target[1]
        except IndexError:
            query = None
        method = event.method.decode('ascii')

        body = b''

        while True:
            event = await self.next_event()
            if type(event) is h11.EndOfMessage:
                break
            body += event.data

        return Request(
            headers=headers,
            method=method,
            query_string=query,
            path=path,
            body=body,
            correlation_id=None
        )

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                await self._read()
            else:
                return event



import json
from collections import defaultdict
from urllib import parse
from .router import Methods
from http import HTTPStatus


class QueryParams:
    """
    A dictionary that stores multiple values per key.

    this has all the normal dictionary methods, and works as normal but does
    not override a key when `add` is used, and also has `getall`
    """
    __slots__ = ['mappings']

    @classmethod
    def from_string(cls, string):
        query_params = QueryParams()
        qs = parse.parse_qsl(string)
        for k, v in qs:
            query_params.add(k, v)

        return query_params

    def __init__(self):
        self.mappings = defaultdict(list)

    def get(self, name, default=None):
        return self.mappings.get(name, [default])[0]

    def getall(self, name, default=None):
        return self.mappings.get(name, default)

    def __getitem__(self, key):
        try:
            return self.mappings.__getitem__(key)[0]
        except IndexError:
            raise KeyError('Invalid Key: {key}'.format(key))

    def __setitem__(self, key, value):
        raise TypeError('MultiDict does not support item assignment. '
                        'Use .add(k, v) instead.')

    def add(self, name, value):
        self.mappings[name].append(value)

    def __str__(self):
        return parse.urlencode(self.mappings, doseq=True)


class Request:
    def __init__(self, headers: dict = None,
                 path: str = None, correlation_id: str = None,
                 method: str = None, query_string: str = None,
                 body: bytes=None, content_type='application/json'):

        if not headers:
            headers = {}
        if not method:
            method = 'GET'
        if isinstance(method, str):
            method = Methods(method.upper())
        self.headers = headers
        self.path = path
        self.correlation_id = correlation_id
        self.method = method
        self.query_string = query_string
        self._query_params = None
        self.body = body
        self.path_params = {}
        self._handler = None
        self.app = None
        self.content_type = content_type
        self._json = None
        self._cookies = None

    @property
    def cookies(self) -> dict:
        if self._cookies is None:
            self._cookies = {}
            for string in self.headers.get('cookie', '').split('; '):
                key, value = string.split('=')
                self._cookies[key] = value

        return self._cookies

    @property
    def query(self) -> QueryParams:
        # parse query string into a dictionary
        if not self._query_params:
            self._query_params = QueryParams.from_string(self.query_string)
        return self._query_params

    def json(self) -> dict:
        if not self._json:
            # convert body into a json dict
            self._json = json.loads(self.body.decode())
        return self._json

    def get_path_var(self, key, default=None):
        return self.path_params.get(key, default=default)

    def __str__(self):
        query = '?' + self.query_string if self.query_string else ''
        return('<Request({method} {path}{query})@{id}>'
               .format(method=self.method, path=self.path,
                       query=query , id=id(self)))


class Response:
    def __init__(self, headers=None, correlation_id=None,
                 body=None, status=HTTPStatus.OK,
                 content_type='application/json', meta: dict=None):
        """
        Response object
        :param headers:
        :param correlation_id:
        :param body: message body
        :param status: status code
        :param content_type:
        :param meta: Extra context information.
            Not to be returned through transport
        """
        if not headers:
            headers = dict()
        if isinstance(status, int):  # convert to enum
            status = HTTPStatus(status)
        if meta is None:
            meta = {}
        self.headers = headers
        self.correlation_id = correlation_id
        self.body = body
        self.status = status
        self.content_type = content_type
        self.meta = meta
        self._data = None
        self._json = None

    def __str__(self):
        return('<Response({status})@{id}>'
               .format(status=self.status, id=id(self)))

    @property
    def data(self):
        if self._data is None and self.body:
            if self.content_type == 'application/json':
                self._data = json.dumps(self.body).encode()
            else:
                self._data = self.body
        return self._data

    def json(self) -> dict:
        """
        Used for client response
        """
        if not self._json:
            # convert body into a json dict
            self._json = json.loads(self.body.decode())
        return self._json


class ResponseError(Exception):
    def __init__(self, status, *, body=None, headers=None,
                 correlation_id=None, reason=None):
        super().__init__()
        if reason and not body:
            body = {'reason': reason}
        self.response = Response(status=status, body=body, headers=headers,
                                 correlation_id=correlation_id)



import traceback
from typing import List, Union, Iterable
import asyncio

from .client import Client
from .webtypes import Request, Response, ResponseError
from .transports.transportabc import TransportABC
from .router import Router


class Application:
    def __init__(self,
                 transport: Union[TransportABC,
                                  Iterable[TransportABC]]=None,
                 *,
                 middlewares: List[callable]=None,
                 default_headers: dict=None,
                 debug: bool=False,
                 router: Router=None):
        if transport is None:
            from .transports.httptransport import HTTPServerTransport
            transport = HTTPServerTransport()
        if not isinstance(transport, tuple):
            transport = (transport,)
        if middlewares is None:
            middlewares = []
        if router is None:
            router = Router()
        if default_headers is None:
            default_headers = {'Server': 'waspy'}
        self.transport = transport
        self.middlewares = middlewares
        self.default_headers = default_headers
        self.debug = debug
        self.router = router
        self.on_start = []
        self._client = None

    @property
    def client(self) -> Client:
        if not self._client:
            self._client = Client(transport=self.transport[0].get_client())
        return self._client

    def run(self):
        loop = asyncio.get_event_loop()
        # wrap handlers in middleware
        loop.run_until_complete(self._wrap_handlers())
        for t in self.transport:
            t.listen(loop=loop)

        # Call on-startup hooks
        for coro in self.on_start:
            loop.run_until_complete(coro(self))

        # todo: fork/add processes?
        for t in self.transport:
            t.start(self.handle_request, loop=loop)
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            print('Shutting down')
        for t in self.transport:
            t.shutdown(loop=loop)

        # todo: Call on-shutdown hooks

        loop.close()

    async def handle_request(self, request: Request) -> Response:
        """
        coroutine: This method is called by Transport
        implementation to handle the actual request.
        It returns a webtype.Response object.
        """
        # ToDo: Send through middlewares
        # response = Response(headers={'hello': 'world'}, body={'foo': 'bar'})
        # Get handler
        try:
            handler = self.router.get_handler_for_request(request)
            response = await handler(request)
            if not isinstance(response, Response):
                if isinstance(response, tuple):
                    body = response[0]
                    status = response[1]
                    response = Response(status=status, body=body)
                elif isinstance(response, dict) or isinstance(response, str):
                    response = Response(body=response)

        except ResponseError as r:
            response = r.response
        except Exception as e:
            traceback.print_exc()
            response = Response(status=500)
        if not response.correlation_id:
            response.correlation_id = request.correlation_id
        # add default headers
        response.headers = {**self.default_headers, **response.headers}

        return response

    async def _wrap_handlers(self):
        handler_gen = self.router._get_and_wrap_routes()

        try:
            handler = next(handler_gen)
            while True:
                wrapped = handler
                for middleware in self.middlewares[::-1]:
                    wrapped = await middleware(self, wrapped)
                handler = handler_gen.send(wrapped)
        except StopIteration:
            pass




import json
from urllib import parse
import uuid

from .webtypes import QueryParams


class Client:
    """ Generic Client class for making a wasp client """
    __slots__ = ('transport',)

    def __init__(self, transport=None, **kwargs):
        if not transport:
            from waspy.transports import HTTPClientTransport
            transport = HTTPClientTransport(**kwargs)
        self.transport = transport

    async def make_request(self, method, service, path, body=None,
                           query_params=None, headers=None,
                           correlation_id=None,
                           content_type='application/json', **kwargs):
        """ Make a request"""
        if content_type == 'application/json' and isinstance(body, dict):
            body = json.dumps(body)
        if isinstance(query_params, dict):
            query_string = parse.urlencode(query_params)
        elif isinstance(query_params, QueryParams):
            query_string = str(QueryParams)
        else:
            query_string = ''
        if not correlation_id:
            correlation_id = uuid.uuid4()
        if isinstance(body, str):
            body = body.encode()
        response = await self.transport.make_request(
            service, method, path, body=body, query=query_string,
            headers=headers, correlation_id=correlation_id,
            content_type=content_type)
        return response

    async def get(self, service, path):
        """ Make a get request """

    async def post(self, service, path, body):
        """ Make a post request """

    async def put(self, service, path, body):
        """ Make a put request """

    async def patch(self, service, path, body):
        """ Make a patche requests """

    async def delete(self, service, path):
        """ Make a delete requests """
